/* cl_nn.c
 * ANDHILL (soccer client for Robocup97)
 * Tomohito Andou <andou@is.titech.ac.jp>
 * Tokyo Institute of Technology
 * Dept of Mathematical and Computing Sciences
 * Copyright (C) 1997 Tomohito Andou
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "cl_nn.h"

#define EPSILON	(1.0)
#define ETA	(0.1)
#define GAMMA	(0.80)

double pa_i[PA_IN+1];
double pa_o[PA_OU];
static double pa_h[PA_HD+1];
static double pa_ih[PA_HD][PA_IN+1] = {
  {-0.5, 0, 0, 0, 0, 2},
  {0, -0.5, 0, 0, 0, 2},
  {0, 0, -2, 0, 0, 4},
  {0, 0, 0, -2, 0, 4},
  {0, 0, 0, 0, 2, -16}
};
static double pa_ho[PA_OU][PA_HD+1] = {
  {-1000, 1000, 400, -500, -1500, 0}
};

double po_i[PO_SET][PO_IN+1];
double po_o[PO_SET][PO_OU];
double po_osum;
double po_io[PO_SET][PO_OU][PO_IN+1];
static double po_io_d[PO_SET][PO_OU][PO_IN+1];

double pass_exec(void)
{
  double sum;
  int i, j;

  pa_i[PA_IN] = 1.0;
  for (j=0; j<PA_HD; j++) {
    sum = 0.0;
    for (i=0; i<PA_IN+1; i++)
      sum += pa_i[i] * pa_ih[j][i];
    pa_h[j] = sigmoid(sum);
  }
  pa_h[PA_HD] = 1.0;
  for (j=0; j<PA_OU; j++) {
    sum = 0.0;
    for (i=0; i<PA_HD+1; i++)
      sum += pa_h[i] * pa_ho[j][i];
    /* pa_o[j] = sigmoid(sum); */
    pa_o[j] = sum;
  }
  return (pa_o[0]);
}

double posi_exec(void)
{
  double sum;
  int i, j, k;

  po_osum = 0.0;
  for (k=0; k<PO_SET; k++) {
    po_i[k][PO_IN] = 1.0;
    for (j=0; j<PO_OU; j++) {
      sum = 0.0;
      for (i=0; i<PO_IN+1; i++)
	sum += po_i[k][i] * po_io[k][j][i];
      po_o[k][j] = sigmoid(sum);
      po_osum += po_o[k][j];
    }
  }
  return (po_osum);
}

/*
 * $B3NN(E*79<PK!(B
 * $B$3$N4X?t$O(Bposi_exec()$B$N8e$K<B9T$9$k$3$H(B
 */
void posi_backprop(int selected)
{
  static double po_odel[PO_SET][PO_OU];
  static double po_io_e[PO_SET][PO_OU][PO_IN+1];
  double dist;
  int i, j, k;

  /* $B8m:95UEAGE(B */
  for (k=0; k<PO_SET; k++)
    for (j=0; j<PO_OU; j++) {
      po_odel[k][j] = (k==selected) ? 1.0 / po_o[k][j] : 0.0;
      po_odel[k][j] -= 1.0 / po_osum;
      po_odel[k][j] *= EPSILON * po_o[k][j] * (1.0 - po_o[k][j]);
      for (i=0; i<PO_IN+1; i++)
	po_io_e[k][j][i] = ETA * po_i[k][i] * po_odel[k][j];
    }

  /* $B3j$i$+$K$7$F(BD$B$KC_$($k(B */
  for (k=0; k<PO_SET; k++) {
    dist = 1.0;
    dist += Pow((k/SMALL_X) - (selected/SMALL_X));
    dist += Pow((k%SMALL_X) - (selected%SMALL_X));
    if (dist > 16) dist = 1000000;
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_IN+1; i++) {
	po_io_e[k][j][i] += po_io_e[selected][j][i] / dist;
	po_io_d[k][j][i] = GAMMA * po_io_d[k][j][i] + po_io_e[k][j][i];
      }
  }
}

void posi_init(void)
{
  int i, j, k;

  for (k=0; k<PO_SET; k++)
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_IN+1; i++)
	po_io_d[k][j][i] = 0.0;
}

/* $B99?7(B */
void posi_reward(double rew)
{
  int i, j, k;

  for (k=0; k<PO_SET; k++)
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_IN+1; i++)
	po_io[k][j][i] += rew * (1.0 - GAMMA) * po_io_d[k][j][i];
}

void posi_rand_init(void)
{
  int i, j, k;

  for (k=0; k<PO_SET; k++)
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_IN+1; i++)
/*	po_io[k][j][i] = 0.0;*/
	po_io[k][j][i] = (random()%101)*0.001 - 0.050;
}

void posi_output(FILE *fp)
{
  int i, j, k;

  for (k=0; k<PO_SET; k++) {
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_IN+1; i++)
	fprintf(fp, "%.2f ", po_io[k][j][i]);
    fprintf(fp, "\n");
  }
}

void posi_input(FILE* fp)
{
  int i, j, k;

  for (k=0; k<PO_SET; k++) {
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_IN+1; i++)
	fscanf(fp, "%lf ", &po_io[k][j][i]);
    fscanf(fp, "\n");
  }
}

double sigmoid(double x)
{
  double xx;

  xx = x * EPSILON;
  if (xx > 6.9)       return (0.999);
  else if (xx < -6.9) return (0.001);
  else return (1.0 / (1.0 + exp(-xx)));
}

/* End Of File */
