/* client.c
 * ANDHILL (soccer client for Robocup97)
 * Tomohito Andou <andou@is.titech.ac.jp>
 * Tokyo Institute of Technology
 * Dept of Mathematical and Computing Sciences
 * Copyright (C) 1997 Tomohito Andou
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>

#include "client.h"
#include "cl_nn.h"

#if 0
#define TEST
#endif

#define MAXMESG		(1000)
#define GO_ANGLE(odist)	(((odist) > 1.0)? 5 + 10/(odist): 15)
#define CHASE_RANGE	(20 + my_stamina/60)
#define KEEPER_RANGE	(20.0)
#define SEE_BALL_FREQ	(6 + ball_dist)
#define SEE_MY_FREQ	(100)
#define SAY_FREQ	(10)
#define REWARD_FREQ	(10)
#define POSI_FREQ	(8)
#define POSI_ROUGH	(6)
#define BALL_ROUGH	(1.0)
#define CHASE_BALL	(BALL_ROUGH*6)
#define VIEW_ROUGH	(5)
#define OBSTACLE_ANGLE	(3)
#define SHOOT_RANGE	(1000)		/* square */
#define MY_KICKABLE	(0.95)
#define ENEMY_KICKABLE	(0.9)
#define BALL_CATCH	(0.50)
#define CATCH_MARGIN	(3.0)
#define PASS_MARGIN	(5.0)

Socket		sock;
int		learn_flag;
int		fixedposi_flag;

long		game_time;
char		game_team[16];
Side		game_side;
int		game_unum;
int		game_mode;

long		my_time;
double		my_x;
double		my_y;
int		my_Sx;
int		my_Sy;
double		my_dir;
int		my_stamina;
int		my_view;	/* 45,90,180 */
double		my_gox;
double		my_goy;

long		my_vel_time;
double		my_vel_x;
double		my_vel_y;
double		my_vel;

long		ball_time;
double		ball_x;
double		ball_y;
int		ball_Sx;
int		ball_Sy;
double		ball_dir;
double		ball_dist;
double		ball_ex;
double		ball_ey;

long		ball_vel_time;
double		ball_vel_x;
double		ball_vel_y;
double		ball_vel;

long		map_time;
char		map_char[SMALL_X][SMALL_Y];
double		sc_base[SMALL_X][SMALL_Y][2];

int		info_flag;
int		turn_flag;
int		seebl_flag;
int		reached_posi;
int		reached_ball;
int		last_say;
int		last_reward;
int		last_posi;

SeenObj		nearest_player;
SeenObj 	seen_player[21];
int		seen_num;
SeenObj		nearest_enemy;
int		seen_enemy;

#if 0
/* andou's position */
double init_posi[11][2] = {
  {-51.5,	0.0},
  {-1.0,	0.0},
  {-39.1,	-8.0},
  {-39.1, 	8.0},
  {-23.0,	-6.5},
  {-23.0,	6.5},
  {-16.4,	0.0},
  {-9.8,	-12.4},
  {-9.8,	12.4},
  {-1.0,	-14.5},
  {-1.0,	14.5}
};
double off_posi[11][2] = {
  {-51.5,	0.0},
  {16.4,	0.0},
  {-39.1, 	-8.0},
  {-39.1, 	8.0},
  {-23.0,	-6.5},
  {-23.0,	6.5},
  {-9.8,	0.0},
  {9.8,		-15.5},
  {9.8,		15.5},
  {36.1,	-12.4},
  {36.1,	12.4}
};
#else
/* muku's position */
double init_posi[11][2] = {
  {-51.5,	0.0},
  {-1.0,	-0.0},
  {-29.5, 	-6.2},
  {-36.1,	6.2},
  {-23.0,	-18.5},
  {-16.4,	18.5},
  {-3.3,	6.2},
  {-1.0,	-18.5},
  {-1.0,	18.5},
  {-1.0,	-12.4},
  {-1.0,	12.4}
};
double off_posi[11][2] = {
  {-51.5,	0.0},
  {3.3,		-6.2},
  {-29.5, 	-6.2},
  {-36.1,	6.2},
  {-23.0,	-18.5},
  {-16.4,	18.5},
  {-3.3,	6.2},
  {16.4,	-18.5},
  {23.0,	18.5},
  {36.1,	-12.4},
  {36.1,	12.4}
};
#endif

/*
 * message $B$NAw<u?.(B
 *
 * $B0z?t(B :
 * fpin : $BF~NO(B($BFbMF$O(B socket $B$KAw$j9~$^$l$k(B)
 * fpout: $B=PNO(B(socket $B$+$i$N<u?.FbMF$,Aw$j9~$^$l$k(B)
 *
 */

void message_loop(FILE* fpin, FILE* fpout)
{
  fd_set		readfds;
  int			in, max_fd, n;
  static char		buf[MAXMESG];
  SensorInfo		info;
  extern void		handler();
  struct timeval	zerotv;
  struct itimerval	itv, oitv;

  zerotv.tv_sec		= 0;
  zerotv.tv_usec	= 0;

  itv.it_interval.tv_sec	= 0;
  itv.it_interval.tv_usec	= 100000;
  itv.it_value.tv_sec		= 0;
  itv.it_value.tv_usec		= 100000;

  signal(SIGALRM, handler);
  setitimer(ITIMER_REAL, &itv, &oitv);

  in = fileno(fpin);
  while (1) {
    sigpause(sigblock(0L));

    signal(SIGALRM, SIG_IGN);

    FD_ZERO(&readfds);
    FD_SET(in, &readfds);
    FD_SET(sock.socketfd, &readfds);

    max_fd = ((in > sock.socketfd) ? in : sock.socketfd) + 1;
    if (select(max_fd, &readfds, 0, 0, &zerotv) == -1) {
      perror("select");
      break;
    }

    if (FD_ISSET(in, &readfds)) {
      if (fgets(buf, MAXMESG, fpin) == NULL) {
	send_message(sock, "(bye)");
	break;
      }
      if (send_message(sock, buf) == FALSE)
	break;
    }

    if (FD_ISSET(sock.socketfd, &readfds))
      while ((n = receive_message(&sock, buf, MAXMESG)) == TRUE) {
#if 0
	fputs(buf, fpout);
#endif
	scan_info(buf, &info, game_team);
	if (info.type == SIT_Hear)
	  hear(info.info.hear);
	else if (info.type == SIT_See)
	  see(info.info.see);
      }
#if 0
    show();
#endif

    action();

#ifdef TEST
    getitimer(ITIMER_REAL, &oitv);
    printf("rest time = %ld / 100000\n", oitv.it_value.tv_usec);
#endif

    signal(SIGALRM, handler);
  }
}

/*
 * $B%?!<%sKh$K>pJs$r99?7(B
 */

void handler(int sig)
{
  double	tx, ty;
  int		i;
  static	long counter = 0;

  /*
   * Users of 4.2 and 4.3BSD systems should un-comment
   * this line, which will make this program
   * behave as if it were on a non-Berkeley system.
   */
  /* signal(sig, SIG_DFL); */

  if (game_mode < 3)  game_time ++;

#ifdef TEST
  printf("\n****** %s %d %ld (%ld) ",
	 game_team, game_unum, game_time, counter++);
#endif

  my_x		+= my_vel_x;
  my_y		+= my_vel_y;
  my_Sx		= L2Sx(my_x);
  my_Sy		= L2Sy(my_y);
  my_stamina	+= STAMINA_INC;
  my_stamina	= restrict(my_stamina, 0, STAMINA_MAX);

  ball_x	+= ball_vel_x;
  ball_y	+= ball_vel_y;
  ball_Sx	= L2Sx(ball_x);
  ball_Sy	= L2Sy(ball_y);
  ball_dir      = Rad2Deg(atan2(ball_y - my_y, ball_x - my_x)) - my_dir;
  ball_dir      = regDeg(ball_dir);
  ball_dist     = sqrt(Pow(ball_x - my_x) + Pow(ball_y - my_y));

  for (i=0; i<seen_num; i++) {
    tx = seen_player[i].dist * cos(Deg2Rad(seen_player[i].dir)) - my_vel;
    ty = seen_player[i].dist * sin(Deg2Rad(seen_player[i].dir));
    seen_player[i].dir = Rad2Deg(atan2(ty, tx));
    seen_player[i].dist = sqrt(Pow(tx) + Pow(ty));
  }

  my_vel_x	*= PL_DECAY;
  my_vel_y	*= PL_DECAY;
  my_vel	*= PL_DECAY;

  ball_vel_x	*= BALL_DECAY;
  ball_vel_y	*= BALL_DECAY;
  ball_vel	*= BALL_DECAY;

#ifdef TEST
  printf("******\n");
#endif
}

/*
 * $B%\!<%kJa3M0LCV$N?dDj(B
 */

void ball_estimate(void)
{
  if (ball_vel > BALL_CATCH &&
      Abs(ball_x) < PITCH_LENGTH/2.0 - CATCH_MARGIN &&
      Abs(ball_y) < PITCH_WIDTH/2.0 - CATCH_MARGIN) {
    ball_ex = (ball_vel_x - BALL_CATCH * ball_vel_x/ball_vel) * BALL_ONESTEP;
    ball_ey = (ball_vel_y - BALL_CATCH * ball_vel_y/ball_vel) * BALL_ONESTEP;
    if (Abs(ball_ex) > 0.001) {
      if (ball_x + ball_ex > PITCH_LENGTH/2.0 - CATCH_MARGIN) {
	ball_ey *= (PITCH_LENGTH/2.0 - CATCH_MARGIN - ball_x) / ball_ex;
	ball_ex = PITCH_LENGTH/2.0 - CATCH_MARGIN - ball_x;
      } else if (ball_x + ball_ex < -PITCH_LENGTH/2.0 + CATCH_MARGIN) {
	ball_ey *= (-PITCH_LENGTH/2.0 + CATCH_MARGIN - ball_x) / ball_ex;
	ball_ex = -PITCH_LENGTH/2.0 + CATCH_MARGIN - ball_x;
      }
    }
    if (Abs(ball_ey) > 0.001) {
      if (ball_y + ball_ey > PITCH_WIDTH/2.0 - CATCH_MARGIN) {
	ball_ex *= (PITCH_WIDTH/2.0 - CATCH_MARGIN - ball_y) / ball_ey;
	ball_ey = PITCH_WIDTH/2.0 - CATCH_MARGIN - ball_y;
      } else if (ball_y + ball_ey < -PITCH_WIDTH/2.0 + CATCH_MARGIN) {
	ball_ex *= (-PITCH_WIDTH/2.0 + CATCH_MARGIN - ball_y) / ball_ey;
	ball_ey = -PITCH_WIDTH/2.0 + CATCH_MARGIN - ball_y;
      }
    }
    ball_ex += ball_x;
    ball_ey += ball_y;
  } else {
    ball_ex = ball_x;
    ball_ey = ball_y;
  }
}

/*
 * $B8D?M$N9TF0$O$3$3$K5-=R(B
 */

void action(void)
{
  /* $B%;%s%5>pJs$,ITB-$7$F$$$J$$$+(B */
  if ((info_flag & 1) == 0) {
    say();
    return;
  } else if ((info_flag & 6) != 6) {
    turn((my_view - VIEW_ROUGH) * turn_flag);
    info_flag &= ~1;
    return;
  } else if (game_time - ball_time > SEE_BALL_FREQ) {
    turn(ball_dir);
    info_flag &= ~3;
    turn_flag = (ball_dir >= 0)? 1: -1;
    return;
  } else if (game_time - my_time > SEE_MY_FREQ) {
    turn(180.0);
    info_flag &= ~5;
    return;
  }

  /* $B%-%C%/%*%UA0$OFCJL$J=hM}(B */
  if (game_mode == 3) {
    if (Pow(my_x-my_gox) + Pow(my_y-my_goy) > Pow(POSI_ROUGH)
	|| my_x <= -PITCH_LENGTH || my_x >= 0
	|| my_y <= -PITCH_WIDTH || my_y >= PITCH_WIDTH) {
      reached_posi = 0;
      move(my_gox, my_goy);
    } else {
      reached_posi = 1;
      if (Abs(ball_dir) > GO_ANGLE(ball_dist))
	turn(ball_dir);
      else
	say();
    }
    return;
  }

  ball_estimate();

  /* $B3X=,$9$k>l9g$O>o$K%M%C%H%o!<%/$rDL$9$3$H(B */
  if (learn_flag == 1)
    position();

  /* $B%\!<%k$,=3$l$k;~(B */
  if (ball_dist < MY_KICKABLE + PL_SIZE + BALL_SIZE) {
    pass();
    if (learn_flag && game_time - last_reward > REWARD_FREQ) {
      calc_reward();
      last_reward = game_time;
    }
  } else {
    if (learn_flag && (game_mode == 0 || game_mode == 1))
      posi_reward(-0.0100);

    /* $B$7$P$i$/(Bsay$B$7$F$J$$;~(B */
    if (game_time - last_say > SAY_FREQ)
      say();

    /* $B%\!<%k$rDI$&(B */
    else if (chase()) {
      if (ball_dist <= CHASE_BALL || ball_vel <= BALL_CATCH)
	go(ball_dir, ball_dist-0.5);
      else if ((reached_ball == 0 &&
		Pow(my_x-ball_ex) + Pow(my_y-ball_ey) > Pow(BALL_ROUGH)) ||
	       (reached_ball == 1 &&
		Pow(my_x-ball_ex) + Pow(my_y-ball_ey) > Pow(POSI_ROUGH))) {
	reached_ball = 0;
	go_abs(ball_ex, ball_ey);
      } else {
	reached_ball = 1;
	turn(ball_dir);
      }
    }

    /* $B%\!<%k$,1s$$;~(B */
    else {
      reached_ball = 0;
      position();
      if ((reached_posi == 0 &&
	   Pow(my_x-my_gox) + Pow(my_y-my_goy) > Pow(POSI_ROUGH/2)) ||
	  (reached_posi == 1 &&
	   Pow(my_x-my_gox) + Pow(my_y-my_goy) > Pow(POSI_ROUGH)))
	reached_posi = 0;
      else
	reached_posi = 1;
      if (reached_posi == 0
	  && !(game_mode == 2 && ball_dist <= 10.0)) {
	go_abs(my_gox, my_goy);
      } else {
	if (Abs(ball_dir) > GO_ANGLE(ball_dist))
	  turn(ball_dir);
	else
	  say();
      }
    }
  }
}

/*
 * $BJs=7$r7W;;$9$k(B
 */

void calc_reward(void)
{
  double	rew;
#if 0
  int		qx, qy;

  /* $BL#J}$+$iN%$l$F$$$k$[$I9bF@E@(B */
  rew = 100.0;
  for (qy=0; qy<SMALL_Y; qy++)
    for (qx=0; qx<SMALL_X; qx++) {
      if ((map_char[qx][qy] & 5) == 5 &&
	  rew > Pow(qx-my_Sx) + Pow(qy-my_Sy))
	rew = Pow(qx-my_Sx) + Pow(qy-my_Sy);
    }
  rew = sqrt(rew) + 0.2;
  rew = 1.0 * rew;
  if (rew > 4.0) rew = 4.0;
#else
  if (seen_num > 0 && nearest_player.dist <= 10)
    rew = 0.4 * nearest_player.dist;
  else
    rew = 4.0;
#endif

  posi_reward(rew);
  printf("%s %2d\treward %g\n", game_team, game_unum, rew);
}

/*
 * $B%]%8%7%g%K%s%0$H$=$N3X=,(B
 */

void position(void)
{
  int		ox, oy, px, py, qx, qy;
  double	max;
  double	dist1, dist2, dist3;

  if (fixedposi_flag) {
    my_gox = off_posi[game_unum-1][0];
    my_goy = off_posi[game_unum-1][1];
    return;
  }
  if (game_time - last_posi <= POSI_FREQ
      || ((game_time%11) != game_unum-1
	  && ((game_time-1)%11) != game_unum-1)) {
    return;
  }
  if (game_unum == 1) {
    my_gox = -51.5;
    my_goy = 0.0;
    return;
  }

  for (py=0; py<SMALL_Y; py++)
    for (px=0; px<SMALL_X; px++) {
      dist1 = dist2 = 100.0;
      for (qy=0; qy<SMALL_Y; qy++)
	for (qx=0; qx<SMALL_X; qx++) {
	  if ((map_char[qx][qy] & 5) == 5 &&
	      dist1 > Pow(qx-px) + Pow(qy-py))
	    dist1 = Pow(qx-px) + Pow(qy-py);
	  else if ((map_char[qx][qy] & 6) == 6 &&
		   dist2 > Pow(qx-px) + Pow(qy-py))
	    dist2 = Pow(qx-px) + Pow(qy-py);
	}
      qx = L2Sx(ball_ex);
      qy = L2Sy(ball_ey);
      dist3 = Pow(qx-px) + Pow(qy-py);
      po_i[py*SMALL_X+px][0] = 1.0 / (dist1 + 1.0);
      po_i[py*SMALL_X+px][1] = 1.0 / (dist2 + 1.0);
      po_i[py*SMALL_X+px][2] = 1.0 / (dist3 + 1.0);
    }
  posi_exec();
  if (learn_flag)
    posi_backprop(my_Sy*SMALL_X+my_Sx);

#if 0
  for (py=0; py<SMALL_Y; py++) {
    for (px=0; px<SMALL_X; px++)
      printf("%.2f ", po_o[py*SMALL_X+px][0]);
    printf("\n");
  }
#endif

  ox = my_Sx;
  oy = my_Sy;
  max = po_o[oy*SMALL_X+ox][0] + (learn_flag? 0.0: 0.0);
  for (py=0; py<SMALL_Y; py++)
    for (px=0; px<SMALL_X; px++)
      if (po_o[py*SMALL_X+px][0] > max
	  && my_stamina - sqrt(Pow(px-my_Sx)+Pow(py-my_Sy))*394 > 500) {
	ox = px;
	oy = py;
	max = po_o[oy*SMALL_X+ox][0];
      }
  my_gox = S2Lx(ox);
  my_goy = S2Ly(oy);
  last_posi = game_time;
}

/*
 * $B=i4|0LCV$r5a$a$k(B
 */

void init_position(void)
{
  int		ox, oy, px, py;
  double	max;

  if (fixedposi_flag) {
    my_gox = init_posi[game_unum-1][0];
    my_goy = init_posi[game_unum-1][1];
    return;
  }
  if (game_unum == 1) {
    my_gox = -51.5;
    my_goy = 0.0;
    return;
  }

  ox = my_Sx;
  oy = my_Sy;
  max = -10000.0;
  for (py=0; py<SMALL_Y; py++)
    for (px=0; px<SMALL_X/2; px++)
      if (po_io[py*SMALL_X+px][0][PO_IN] > max) {
	ox = px;
	oy = py;
	max = po_io[oy*SMALL_X+ox][0][PO_IN];
      }
  my_gox = S2Lx(ox);
  my_goy = S2Ly(oy);
}

/*
 * $B%\!<%k$rDI$&$+$I$&$+$NH=CG(B
 */

int chase(void)
{
  double ll, mm;
  int i;

  if (game_mode == 2) return 0;

  if (ball_dist <= CHASE_BALL) return 1;

  /* $B1s$$$H$-$ODI$o$J$$(B */
  if (sqrt(Pow(ball_ex-my_x) + Pow(ball_ey-my_y)) > CHASE_RANGE
      && !(game_mode == 1 && ball_dist <= CHASE_RANGE*2))
    return 0;

  /* $B%-!<%Q!<$OFCJL(B */
  if (game_unum == 1) {
    if (ball_ex > - PITCH_LENGTH/2.0 + KEEPER_RANGE
	|| Abs(ball_ey) > KEEPER_RANGE)
      return 0;
  }

  /* $B%\!<%k$N6a$/$KL#J}$,$$$k;~$ODI$o$J$$(B */
  for (i=0; i<seen_num; i++) {
    if (seen_player[i].id.player.team == WT_Our) {
      mm = (seen_player[i].dist
	     * cos(Deg2Rad(seen_player[i].dir + my_dir)) + my_x
	     - ball_ex);
      ll = Pow(mm);
      mm = (seen_player[i].dist
	     * sin(Deg2Rad(seen_player[i].dir + my_dir)) + my_y
	     - ball_ey);
      ll += Pow(mm);
      ll = sqrt(ll);
      mm = sqrt(Pow(my_x - ball_ex) + Pow(my_y - ball_ey));
      if (ll < mm*0.4 || ll < mm-5.0)
	return 0;
    }
  }
  return 1;
}

/*
 * $B$"$kCOE@$K0\F0(B
 */

void go_abs(double ox, double oy)
{
  double odir, odist;

  odir = Rad2Deg(atan2(oy - my_y, ox - my_x)) - my_dir;
  odir = regDeg(odir);
  odist = sqrt(Pow(ox - my_x) + Pow(oy - my_y));
  go(odir, odist);
}

/*
 * $B$"$kJ}8~$K0\F0(B
 */

void go(double odir, double odist)
{
  double	pdir, ddir;

  /* $B:G$b6a$$%W%l%$%d$rHr$1$k(B */
  if (seen_num > 0 && nearest_player.dist < odist) {
    if (PL_SIZE * 2 > nearest_player.dist)
      nearest_player.dist = PL_SIZE * 2;
    pdir = Rad2Deg(asin(PL_SIZE * 2 / nearest_player.dist));
    ddir = regDeg(odir - nearest_player.dir);
    if (ddir <= 0 && ddir > - pdir - GO_ANGLE(odist))
      odir = regDeg(nearest_player.dir - pdir - GO_ANGLE(odist));
    else if (ddir >= 0 && ddir < pdir + GO_ANGLE(odist))
      odir = regDeg(nearest_player.dir + pdir + GO_ANGLE(odist));
  }

  if (Abs(odir) > GO_ANGLE(odist))
    turn(odir);
  else
    dash((odist-my_vel)*POWER_MAX / PL_ONESTEP);
}

/* 
 * $B%\!<%k$r$I$3$K=3$k$+H=CG$9$k(B
 */

void pass(void)
{
  int		tmp_char;
  int		ox, oy, px, py, qx, qy;
  double	max;

  /* $B%4!<%k$KFO$/;~(B */
  if ((Pow(ball_ex - PITCH_LENGTH/2.0) + Pow(ball_ey) < SHOOT_RANGE) ||
      (Pow(ball_x - PITCH_LENGTH/2.0) + Pow(ball_y) < SHOOT_RANGE)) {
    kick(PITCH_LENGTH/2.0, 0.0);
    return;
  }

  /* $B%9%b!<%k%F!<%V%k$+$i$NF0E*I>2A(B */
  tmp_char = map_char[my_Sx][my_Sy];
  map_char[my_Sx][my_Sy] |= 5;
#ifdef TEST
  show();
#endif
  max = -100;
  ox = SMALL_X;
  oy = SMALL_Y/2;
  for (py=0; py<SMALL_Y; py++) {
    for (px=0; px<SMALL_X; px++) {
      pa_i[0] = sc_base[px][py][0];
      pa_i[1] = sc_base[px][py][1];
      pa_i[2] = 100;
      pa_i[3] = 100;
      for (qy=0; qy<SMALL_Y; qy++)
	for (qx=0; qx<SMALL_X; qx++) {
	  if ((map_char[qx][qy] & 5) == 5 &&
	      pa_i[2] > Pow(qx-px) + Pow(qy-py)) {
	    pa_i[2] = Pow(qx-px) + Pow(qy-py);
	  } else if ((map_char[qx][qy] & 6) == 6 &&
		     pa_i[3] > Pow(qx-px) + Pow(qy-py)) {
	    pa_i[3] = Pow(qx-px) + Pow(qy-py);
	  }
	}
      pa_i[2] = sqrt(pa_i[2]);
      pa_i[3] = sqrt(pa_i[3]);
      pa_i[4] = sqrt(Pow(px-ball_Sx) + Pow(py-ball_Sy));
      pass_exec();
      if (pa_o[0] > max) {
	ox = px;
	oy = py;
	max = pa_o[0];
      }
#ifdef TEST
      printf("%5d", (int) pa_o[0]);
#endif
    }
#ifdef TEST
    printf("\n");
#endif
  }
  map_char[my_Sx][my_Sy] = tmp_char;

  if (ox == my_Sx && oy == my_Sy) {
    /* $B%\!<%k$r%-!<%W(B */
    turn(my_view - VIEW_ROUGH);
    info_flag &= ~1;
  }
  else
    /* $B%Q%9(B */
    kick(S2Lx(ox), S2Ly(oy));
}

/*
 * $B@)Ls$r2r>C$7$F=3$k(B($B4pAC(B)
 */

void kick(double kick_x, double kick_y)
{
  double	kick_dir, kick_dist, kick_pow, kick_scope;
  int		kick_flag, uncertain, dribble_flag, shoot_flag, i;
  double	xx, yy, ll, mm;

  kick_flag	= 0;
  dribble_flag	= 0;
  shoot_flag	= 0;
  uncertain	= 0;

  /* $B%7%e!<%H$+$I$&$+(B */
  if (kick_x >= PITCH_LENGTH/2.0 && Abs(kick_y) < GOAL_WIDTH/2.0) {
    kick_dist = sqrt(Pow(kick_y - ball_y) + Pow(kick_x - ball_x));
    if (kick_dist < 1.0) kick_dist = 1.0;
    kick_x += (kick_x - ball_x) * 100 / kick_dist;
    kick_y += (kick_y - ball_y) * 100 / kick_dist;
    shoot_flag = 1;
  }

  /* $B2>$NJ}8~!"5wN%$r7W;;(B */
  kick_dir = Rad2Deg(atan2(kick_y - ball_y, kick_x - ball_x)) - my_dir;
  kick_dir = regDeg(kick_dir);
  kick_dist = sqrt(Pow(kick_y - ball_y) + Pow(kick_x - ball_x));

  /* $B8e$m$K=3$k(B */
  if (PL_SIZE+BALL_SIZE+0.05 > ball_dist)
    ball_dist = PL_SIZE+BALL_SIZE+0.05;
  kick_scope = 180 - Rad2Deg(asin((PL_SIZE+BALL_SIZE+0.05) / ball_dist));

  if (Abs(regDeg(kick_dir - ball_dir)) > kick_scope) {
    /* $B:G$b6a$$E($N$$$J$$J}8~$K2s$k(B */
    if (seen_enemy > 0 &&
	nearest_enemy.dist <= 5.0) {
      if (regDeg(nearest_enemy.dir - ball_dir) >= 0)
	kick_flag = 1;
      else
	kick_flag = -1;
    }
    if (kick_flag == 0) {
      if (regDeg(kick_dir - ball_dir) >= 0)
	kick_flag = -1;
      else
	kick_flag = 1;
    }

    kick_dir = ball_dir - kick_scope * kick_flag;
    kick_dir = regDeg(kick_dir);
    /*kick_dist = 2 * ball_dist*cos(Deg2Rad(180 - kick_scope));*/
    kick_dist = 5;
    dribble_flag = 1;
    shoot_flag = 0;
  }

  /* $BE($N$$$J$$J}8~$K=$@5$9$k(B */
  for (i=0; i<seen_num; i++)
    if (seen_player[i].dist < kick_dist) {

      ll = sqrt(Pow(seen_player[i].dist) + Pow(ball_dist)
		- 2 * seen_player[i].dist * ball_dist
		* cos(Deg2Rad(seen_player[i].dir - ball_dir)));
      if (ll < PL_SIZE + BALL_SIZE + ENEMY_KICKABLE) {
	ll = PL_SIZE + BALL_SIZE + ENEMY_KICKABLE;
	uncertain = 1;
      }
      if (ll < (mm = sin(Deg2Rad(seen_player[i].dir - ball_dir))
		* seen_player[i].dist))
	ll = mm;
      yy = Rad2Deg(asin((PL_SIZE + BALL_SIZE + ENEMY_KICKABLE) / ll));
      xx = Rad2Deg(asin(mm / ll));

      if (Abs(regDeg(ball_dir + xx - kick_dir)) < yy + OBSTACLE_ANGLE) {
	if (kick_flag == 0) {
	  if (regDeg(ball_dir + xx - kick_dir) >= 0)
	    kick_flag = -1;
	  else
	    kick_flag = 1;
	}
	kick_dir = ball_dir + xx + (yy + OBSTACLE_ANGLE) * kick_flag;
	kick_dir = regDeg(kick_dir);
      }
    }

  /* $B=3$kJ}8~!"5wN%$r5a$a$k(B */
  kick_x = ball_x + kick_dist * cos(Deg2Rad(my_dir + kick_dir));
  kick_y = ball_y + kick_dist * sin(Deg2Rad(my_dir + kick_dir));
  if (dribble_flag == 0 && shoot_flag == 0) {
    kick_x = restrict(kick_x, -PITCH_LENGTH/2.0 + PASS_MARGIN,
		      PITCH_LENGTH/2.0 - PASS_MARGIN);
    kick_y = restrict(kick_y, -PITCH_WIDTH/2.0 + PASS_MARGIN,
		      PITCH_WIDTH/2.0 - PASS_MARGIN);
  }

  kick_dir = Rad2Deg(atan2(kick_y - ball_y - ball_vel_y*BALL_ONESTEP,
			   kick_x - ball_x - ball_vel_x*BALL_ONESTEP))
    - my_dir;
  kick_dir = regDeg(kick_dir);
  kick_dist = sqrt(Pow(kick_y - ball_y - ball_vel_y*BALL_ONESTEP) +
		   Pow(kick_x - ball_x - ball_vel_x*BALL_ONESTEP));

  /* $B=3$kNO$r5a$a$k(B */
  kick_pow = kick_dist * 100.0 / BALL_DISTANCE;
  kick_pow = restrict(kick_pow, POWER_MIN, POWER_MAX);

  send_com_kick(&sock, kick_pow, kick_dir);

  if (uncertain == 0) {
    ball_vel_x	+= kick_pow * KICK_RATE * cos(Deg2Rad(kick_dir + my_dir));
    ball_vel_y	+= kick_pow * KICK_RATE * sin(Deg2Rad(kick_dir + my_dir));
    ball_vel	= sqrt(Pow(ball_vel_x) + Pow(ball_vel_y));
  }
}

/*
 * $BA0?J$9$k(B($B4pAC(B)
 */

void dash(double dash_pow)
{
  dash_pow = restrict(dash_pow, POWER_MIN, POWER_MAX);
  send_com_dash(&sock, dash_pow);

  if (dash_pow > my_stamina)
    dash_pow = my_stamina;
  my_stamina -= dash_pow;

  my_vel_x	+= dash_pow * DASH_RATE * cos(Deg2Rad(my_dir));
  my_vel_y	+= dash_pow * DASH_RATE * sin(Deg2Rad(my_dir));
  my_vel	= sqrt(Pow(my_vel_x) + Pow(my_vel_y));
}

/*
 * $B8~$-$rJQ$($k(B($B4pAC(B)
 */

void turn(double turn_dir)
{
  double tmp1, tmp2;
  int i;

  turn_dir = regDeg(turn_dir);
  tmp1 = turn_dir * (1.0 + 5.0 * my_vel);
  tmp2 = restrict(tmp1, -180, 180);
  if (tmp1 != 0) turn_dir *= tmp2 / tmp1;

  send_com_turn(&sock, tmp2);

  my_dir += turn_dir;
  my_dir = regDeg(my_dir);
  ball_dir -= turn_dir;
  ball_dir = regDeg(ball_dir);
  for (i=0; i<seen_num; i++) {
    seen_player[i].dir -= turn_dir;
    seen_player[i].dir = regDeg(seen_player[i].dir);
  }
}

/*
 * $B=V4V0\F0(B($B4pAC(B)
 */

void move(double x, double y)
{
  int i, j;

  send_com_move(&sock, x, y);

  my_x		= x;
  my_y		= y;
  my_vel_x	= 0.0;
  my_vel_y	= 0.0;
  my_vel	= 0.0;

  ball_time	= game_time;
  ball_x	= 0.0;
  ball_y	= 0.0;

  ball_vel_time	= game_time;
  ball_vel_x	= 0.0;
  ball_vel_y	= 0.0;
  ball_vel	= 0.0;

  for (j=0; j<SMALL_Y; j++)
    for (i=0; i<SMALL_X; i++)
      map_char[i][j] = 0;

  info_flag	= 2;
  turn_flag	= 1;
  seebl_flag	= 0;
  reached_posi	= 0;
  reached_ball	= 0;
  last_say	= -999;
  last_reward	= -999;
  last_posi	= -999;

  seen_num	= 0;
  seen_enemy	= 0;

  posi_init();
}

/*
 * $B%l%U%j!<$+$i$N>pJs$rJ9$-<h$k(B
 */

void get_mode(char* message)
{
  char	kick_side;
  FILE	*fp;
  char	fname[20];

  if (strncmp("play_on", message, 7) == 0)
    game_mode = 0;
  else if (sscanf(message, "kick_off_%c", &kick_side) == 1 ||
	   sscanf(message, "kick_in_%c", &kick_side) == 1 ||
	   sscanf(message, "free_kick_%c", &kick_side) == 1 ||
	   sscanf(message, "corner_kick_%c", &kick_side) == 1 ||
	   sscanf(message, "goal_kick_%c", &kick_side) == 1) {
    if ((kick_side == 'l' && game_side == S_LEFT) ||
	(kick_side == 'r' && game_side == S_RIGHT))
      game_mode = 1;
    else
      game_mode = 2;
  } else if (strncmp("goal_", message, 5) == 0 ||
	     strncmp("before_kick_off", message, 15) == 0) {
    if (learn_flag && sscanf(message, "goal_%c", &kick_side) == 1) {
      if ((kick_side == 'l' && game_side == S_LEFT) ||
	  (kick_side == 'r' && game_side == S_RIGHT)) {
	posi_reward(1.0);
	printf("%s %2d\treward %g\n", game_team, game_unum, 1.0);
      } else {
	posi_reward(-1.0);
	printf("%s %2d\treward %g\n", game_team, game_unum, -1.0);
      }
    }
    game_mode = 3;
    init_position();
    move(my_gox, my_goy);
    /* $B3X=,ESCf7k2L$r=PNO$9$k(B */
    if (learn_flag) {
      sprintf(fname, "%s%02d.pos", game_team, game_unum);
      if ((fp = fopen(fname, "w")) == NULL) {
	fprintf(stderr, "%s: can't write.\n", fname);
	exit(1);
      }
      posi_output(fp);
      fclose(fp);
    }
  } else if (strncmp("time_over", message, 9) == 0)
    game_mode = 4;
  else /* unspecified */
    game_mode = 0;
}

/*
 * $B%]%8%7%g%s$J$I$rEAC#$9$k(B
 */

void say(void)
{
  static char	buf[MAXMESG];
  static int	last_Sx, last_Sy;
  int		x, y;
  int		tmp_char;

  /* $BEAC#$9$k;~$K$O<+J,$N0LCV$b4^$a$k(B */
  tmp_char = map_char[my_Sx][my_Sy];
  if (my_time >= map_time)
    map_char[my_Sx][my_Sy] |= 0xd;

  /* $BA02s<+J,$,$$$?>l=j$K$O<+J,$,$$$J$$(B */
  map_char[last_Sx][last_Sy] |= 8;

  sprintf(buf, "%-11s %04ld %02d ", game_team, game_time%10000, game_mode);
  sprintf(buf+20, "%04ld %04d %04d ",
	  ball_time%10000, (int)(ball_x * 10), (int)(ball_y * 10));
  sprintf(buf+35, "%04ld %04d %04d ",
	  ball_vel_time%10000, (int)(ball_vel_x * 100), (int)(ball_vel_y * 100));
  sprintf(buf+50, "%04ld ", map_time%10000);
  for (y=0; y<SMALL_Y; y++)
    for (x=0; x<SMALL_X; x++)
      sprintf(buf+55+y*SMALL_X+x, "%x", map_char[x][y]);
  sprintf(buf+55+SMALL_X*SMALL_Y, "\0");

  map_char[my_Sx][my_Sy] = tmp_char;

  last_Sx = my_Sx;
  last_Sy = my_Sy;
  last_say = game_time;
  send_com_say(&sock, buf);
}

/*
 * $B%]%8%7%g%s$J$I$rJ9$-<h$k(B
 */

void hear(HearInfo info)
{
  static char	name[16];
  int		x, y;
  long		t;
  char		c;
  int		tmp_char;

  game_time = info.time;

  if (info.sender == ST_Self) return;

  if (info.sender == ST_Referee) {
    get_mode(info.message);
    return;
  }

  /* return; */

  /* $BL#J}$N%a%C%;!<%8(B */
  if (strlen(info.message) != 231) return;
  if (sscanf(info.message, "%s ", name) == 0) return;
  if (strcmp(name, game_team) != 0) return;

  /* game_time and game_mode are unused */

  /* $B%\!<%k$N>pJs(B */
  sscanf(info.message+20, "%ld ", &t);
  if (t > 9000 && (game_time % 10000) < 1000)
    t += (game_time / 10000) * 10000 - 10000;
  else
    t += (game_time / 10000) * 10000;
  if (t > ball_time + 2*seebl_flag) {
    sscanf(info.message+25, "%d %d ", &x, &y);
    info_flag	|= 2;
    seebl_flag	= 0;
    ball_x	= x * 0.1;
    ball_y	= y * 0.1;
    ball_Sx	= L2Sx(ball_x);
    ball_Sy	= L2Sy(ball_y);
    ball_dir	= Rad2Deg(atan2(ball_y - my_y, ball_x - my_x)) - my_dir;
    ball_dir	= regDeg(ball_dir);
    ball_dist	= sqrt(Pow(ball_x - my_x) + Pow(ball_y - my_y));
    ball_time	= t;
  }

  /* $B%\!<%k$NB.EY$N>pJs(B */
  sscanf(info.message+35, "%ld ", &t);
  if (t > 9000 && (game_time % 10000) < 1000)
    t += (game_time / 10000) * 10000 - 10000;
  else
    t += (game_time / 10000) * 10000;
  if (t > ball_vel_time + 2*seebl_flag) {
    sscanf(info.message+40, "%d %d ", &x, &y);
    ball_vel_x		= x * 0.01;
    ball_vel_y		= y * 0.01;
    ball_vel		= sqrt(Pow(ball_vel_x) + Pow(ball_vel_y));
    ball_vel_time	= t;
  }

  /* $B%]%8%7%g%s$N>pJs(B */
  sscanf(info.message+50, "%ld ", &t);
  if (t > 9000 && (game_time % 10000) < 1000)
    t += (game_time / 10000) * 10000 - 10000;
  else
    t += (game_time / 10000) * 10000;
  if (t >= map_time - 1) {
    tmp_char = map_char[my_Sx][my_Sy];
    for (y=0; y<SMALL_Y; y++)
      for (x=0; x<SMALL_X; x++) {
	sscanf(info.message+55+y*SMALL_X+x, "%c", &c);
	c = (c>='0'&&c<='9') ? (c-'0') : (10+c-'a');
	if (c & 8) {
	  c &= ~8;
	  if ((c & 4) && (map_char[x][y] & 4))
	    map_char[x][y] |= c;
	  else if ((map_char[x][y] & 8) == 0)
	    map_char[x][y] = c;
	}
      }
    map_char[my_Sx][my_Sy] = tmp_char;
    /*map_time = t;*/
  }
}

/*
 * $B;k3P>pJs$rFI$_$H$k(B
 */

void see(SeeInfo info)
{
  PosState	pstate;
  double	tmp_dir, tmp_x, tmp_y;
  int		x, y, i;
  int		out_of_field;

  info_flag |= 1;
  game_time = info.time;

  /* $B8=:_0LCV$r5a$a$k(B */
  if (estimate_current_pos(&info, game_side, &pstate)) {
    info_flag |= 4;
    my_x = pstate.pos.x;
    my_y = pstate.pos.y;
    my_Sx = L2Sx(my_x);
    my_Sy = L2Sy(my_y);
    my_dir = pstate.dir;
    my_time = info.time;
  } else { /* $B3QEY$@$1$G$b5a$a$k(B */
    out_of_field = 0;
    for (i=0; i<info.n_obj; i++) {
      if (info.obj[i].type == OT_Line) {
	if (i != info.n_obj - 1 && info.obj[i+1].type == OT_Line)
	  out_of_field = 1;
	switch(info.obj[i].id.lpos) {
	 case LP_T:
	  if (out_of_field)
	    my_dir = - info.obj[i].dir - ((info.obj[i].dir > 0) ? -180 : 0);
	  else
	    my_dir = - info.obj[i].dir - ((info.obj[i].dir < 0) ? 180 : 0);
	  break;
	 case LP_B:
	  if (out_of_field)
	    my_dir = - info.obj[i].dir - ((info.obj[i].dir < 0) ? 180 : 0);
	  else
	    my_dir = - info.obj[i].dir - ((info.obj[i].dir > 0) ? -180 : 0);
	  break;
	 case LP_L:
	  if (out_of_field)
	    my_dir = - info.obj[i].dir - ((info.obj[i].dir > 0) ? -90 : 90);
	  else
	    my_dir = - info.obj[i].dir - ((info.obj[i].dir < 0) ? -90 : 90);
	  break;
	 case LP_R:
	  if (out_of_field)
	    my_dir = - info.obj[i].dir - ((info.obj[i].dir < 0) ? -90 : 90);
	  else
	    my_dir = - info.obj[i].dir - ((info.obj[i].dir > 0) ? -90 : 90);
	  break;
	 case LP_Unknown:
	  break;
	}
	if (game_side == S_RIGHT)
	  my_dir += ((my_dir > 0) ? -180 : 180);
	break;
      }
    }
    if (i == info.n_obj) info_flag &= ~4;
  }

  /* $B%9%b!<%k%F!<%V%k$OL$99?7>uBV(B */
  if (info.time > map_time)
    for (y=0; y<SMALL_Y; y++)
      for (x=0; x<SMALL_X; x++)
	map_char[x][y] &= ~8;

  /* $B<+J,$NB.EY$r5a$a$k(B */
  for (i=0; i<info.n_obj; i++) {
    if ((info.obj[i].type == OT_Flag || info.obj[i].type == OT_Goal) &&
	Abs(info.obj[i].dir_chng) <= 180 &&
	Abs(info.obj[i].dist_chng) < 8.0) {
      tmp_x = info.obj[i].dist_chng * cos(Deg2Rad(info.obj[i].dir))
	+ Deg2Rad(info.obj[i].dir_chng) *
	  info.obj[i].dist * sin(Deg2Rad(info.obj[i].dir));
      tmp_y = info.obj[i].dist_chng * sin(Deg2Rad(info.obj[i].dir))
	+ Deg2Rad(info.obj[i].dir_chng) *
	  info.obj[i].dist * cos(Deg2Rad(info.obj[i].dir));
      my_vel_x	= - tmp_x*cos(Deg2Rad(my_dir)) + tmp_y*sin(Deg2Rad(my_dir));
      my_vel_y	= - tmp_y*cos(Deg2Rad(my_dir)) - tmp_x*sin(Deg2Rad(my_dir));
      my_vel	= sqrt(Pow(my_vel_x) + Pow(my_vel_y));
      my_vel_time	= info.time;
      break;
    }
  }

  seen_num = 0;
  seen_enemy = 0;
  for (i=0; i<info.n_obj; i++) {

    /* $B%W%l%$%d!<$N>pJs$r%9%b!<%k%F!<%V%k$KG[CV(B */
    if (info.obj[i].type == OT_Player) {
      x = L2Sx(my_x+info.obj[i].dist*cos(Deg2Rad(info.obj[i].dir+my_dir)));
      y = L2Sy(my_y+info.obj[i].dist*sin(Deg2Rad(info.obj[i].dir+my_dir)));
      map_char[x][y] |= 0xc;
      if (info.obj[i].id.player.team == WT_Our)
	map_char[x][y] |= 1;
      else if (info.obj[i].id.player.team == WT_Opponent)
	map_char[x][y] |= 2;
      /*if (info.obj[i].dist < 20 && info.obj[i].id.player.team != WT_Our) {*/
      /*if (info.obj[i].id.player.team != WT_Our) {*/
      if (seen_num == 0)
	nearest_player = info.obj[i];
      else if (info.obj[i].dist < nearest_player.dist)
	nearest_player = info.obj[i];
      seen_player[seen_num] = info.obj[i];
      seen_num ++;
      if (info.obj[i].id.player.team != WT_Our) {
	if (seen_enemy == 0)
	  nearest_enemy = info.obj[i];
	else if (info.obj[i].dist < nearest_enemy.dist)
	  nearest_enemy = info.obj[i];
	seen_enemy ++;
      }
    }

    if (info.obj[i].type == OT_Ball && game_mode != 3) {
      ball_dir	= info.obj[i].dir;
      ball_dist	= info.obj[i].dist;

      /* $B%\!<%k$N0LCV$rGD0.(B */
      info_flag |= 2;
      seebl_flag= 1;
      ball_x	= my_x + ball_dist * cos(Deg2Rad(ball_dir+my_dir));
      ball_y	= my_y + ball_dist * sin(Deg2Rad(ball_dir+my_dir));
      ball_Sx	= L2Sx(ball_x);
      ball_Sy	= L2Sy(ball_y);
      ball_time	= info.time;

      /* $B%\!<%k$NB.EY(B */
      if (Abs(info.obj[i].dir_chng) <= 180 &&
	  Abs(info.obj[i].dist_chng) < 6.0) {
	tmp_x	= info.obj[i].dist_chng * cos(Deg2Rad(ball_dir))
	  + Deg2Rad(info.obj[i].dir_chng)* ball_dist* sin(Deg2Rad(ball_dir));
	tmp_y	= info.obj[i].dist_chng * sin(Deg2Rad(ball_dir))
	  + Deg2Rad(info.obj[i].dir_chng)* ball_dist* cos(Deg2Rad(ball_dir));
	ball_vel_x = tmp_x*cos(Deg2Rad(my_dir)) - tmp_y*sin(Deg2Rad(my_dir))
	  + my_vel_x;
	ball_vel_y = tmp_y*cos(Deg2Rad(my_dir)) + tmp_x*sin(Deg2Rad(my_dir))
	  + my_vel_y;
	ball_vel	= sqrt(Pow(ball_vel_x) + Pow(ball_vel_y));
	ball_vel_time	= info.time;
      } else {
	ball_vel_x	= 0.0;
	ball_vel_y	= 0.0;
	ball_vel	= 0.0;
      }
    }
  }
  /* $B%\!<%k$,8+$($k$O$:$J$N$K8+$($J$$;~(B */
  if (ball_time != info.time
      && (info_flag & 2) == 2
      && Abs(ball_dir) <= my_view/2.0 - VIEW_ROUGH*3.0) {
    info_flag &= ~2;
    seebl_flag = 0;
    turn_flag = (ball_dir >= 0)? 1: -1;
  }

  /* $B%9%b!<%k%F!<%V%k$K6uGr$rKd$a$k(B */
  tmp_x = my_x + cos(Deg2Rad(my_dir)) * 3.0 / sin(Deg2Rad(my_view/2.0));
  tmp_y = my_y + sin(Deg2Rad(my_dir)) * 3.0 / sin(Deg2Rad(my_view/2.0));
  for (y=0; y<SMALL_Y; y++)
    for (x=0; x<SMALL_X; x++) {
      tmp_dir = Rad2Deg(atan2(S2Ly(y) - tmp_y, S2Lx(x) - tmp_x)) - my_dir;
      tmp_dir = regDeg(tmp_dir);
      if (Abs(tmp_dir) < my_view/2.0 && (map_char[x][y] & 8) == 0)
	map_char[x][y] = 8;
    }

  map_char[my_Sx][my_Sy] |= 8;
  map_time = info.time;
}

/*
 * $BFbIt>uBV$rI=<($9$k(B($B%F%9%HMQ(B)
 */

void show(void)
{
  int x, y;

  printf("game\t: %06ld %s side %d unum %d mode %d\n",
	 game_time, game_team, game_side, game_unum, game_mode);
  printf("my\t: %06ld px %g py %g dr %g st %d vw %d\n",
	 my_time, my_x, my_y, my_dir, my_stamina, my_view);
  printf("my_v\t: %06ld vx %g vy %g\n",
	 my_vel_time, my_vel_x, my_vel_y);
  printf("ball\t: %06ld px %g py %g dr %g ds %g\n",
	 ball_time, ball_x, ball_y, ball_dir, ball_dist);
  printf("ball_v\t: %06ld vx %g vy %g\n",
	 ball_vel_time, ball_vel_x, ball_vel_y);
  printf("misc\t: if %d sb %d sn %d se %d\n",
	 info_flag, seebl_flag, seen_num, seen_enemy);

  printf("my_S\t: %d %d\n", my_Sx, my_Sy);
  printf("ball_S\t: %d %d\n", ball_Sx, ball_Sy);

  printf("pos\t: %06ld\n", map_time);
  for (y=0; y<SMALL_Y; y++) {
    for (x=0; x<SMALL_X; x++)
      printf("%x ", map_char[x][y]);
    printf("\n");
  }

  fflush(stdout);
}

/*
 * $BFbIt>uBV$r=i4|2=$9$k(B
 */

void initialize(void)
{
  int		x, y;

  my_time	= -999;
  my_x		= -999.0;
  my_y		= -999.0;
  my_Sx		= 0;
  my_Sy		= 0;
  my_dir	= 0.0;
  my_stamina	= STAMINA_MAX;
#if 1
  send_com_change_view(&sock, VW_Narrow, VQ_high);
  my_view	= 45;
#else
  send_com_change_view(&sock, VW_Normal, VQ_high);
  my_view	= 90;
#endif
  usleep(WAIT_TIME);

  my_vel_time	= -999;
  my_vel_x	= 0.0;
  my_vel_y	= 0.0;
  my_vel	= 0.0;

  ball_time	= -999;
  ball_x	= 0.0;
  ball_y	= 0.0;
  ball_Sx	= 0;
  ball_Sy	= 0;
  ball_dir	= 0.0;
  ball_dist	= 0.0;

  ball_vel_time	= -999;
  ball_vel_x	= 0.0;
  ball_vel_y	= 0.0;
  ball_vel	= 0.0;

  map_time	= -999;
  for (y=0; y<SMALL_Y; y++)
    for (x=0; x<SMALL_X; x++) {
      map_char[x][y] = 0;
      sc_base[x][y][0] = sqrt(Pow(x) +  Pow(y-(SMALL_Y-1)/2));
      sc_base[x][y][1] = sqrt(Pow(x-(SMALL_X-1)) + Pow(y-(SMALL_Y-1)/2));
    }

  info_flag	= 0;
  turn_flag	= 1;
  seebl_flag	= 0;
  reached_posi	= 0;
  reached_ball	= 0;
  last_say	= -999;
  last_reward	= -999;
  last_posi	= -999;

  seen_num	= 0;
  seen_enemy	= 0;

  posi_init();
}

/* End Of File */
